/***********************************************************************/
/*  Easyflow is (C) 1999-2001 lamagra.seKure.de  This program is free  */
/*  software; you can redistribute it and/or modify it under the terms */
/*  of the GNU General Public License as published by the Free Software*/
/*  Foundation; Version 2.  This guarantees your right to use, modify, */
/*  and redistribute this software under certain conditions.           */
/*                                                                     */
/*  If you received these files with a written license agreement       */
/*  stating terms other than the (GPL) terms above, then that          */
/*  alternative license agreement takes precendence over this comment. */
/*                                                                     */
/*  Source is provided to this software because we believe users have  */
/*  a right to know exactly what a program is going to do before they  */
/*  run it.  This also allows you to audit the software for security   */
/*  holes (none have been found so far).                               */
/*                                                                     */
/*  This program is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of     */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  */
/*  General Public License for more details (                          */
/*  http://www.gnu.org/copyleft/gpl.html ).                            */
/*                                                                     */
/***********************************************************************/
#include "lstring.hpp"
#include <ctype.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

void DoSequence(unsigned long *ret);
int sizeof_sc(char *);

char statement[2048];
int idx = 0, len, level = 0;

/* tokens */
enum token_t {DIGIT, OPERATOR, END, ERROR, VARIABLE} tokentype;
char *token;
lstring tokenstring;

int issep(int ch)
{
    static const char seps[] = "+-/*%^()<>";
    int i;
    
    for(i = 0; i < 10; i++)
    {
        if(ch == seps[i])
            return 1;
    }
    
    return 0;
}

const char hexlist[] = "0123456789abcdefx";

int ishex(int ch)
{
    int i;

    for(i = 0; i < 17; i++)
    {
        if(ch == hexlist[i])
            return 1;
    }
    
    return 0;
}

void nexttoken()
{
   tokenstring.clear();

   while(isspace(statement[idx]) && idx < len)
        idx++;

   if(statement[idx] == 0 || statement[idx] == '\n')
   {
        tokentype = END;
        return;
   }

   if(issep(statement[idx]))
   {
        tokentype = OPERATOR;

        if(statement[idx] == ')' && level == 0)
        {
            cout << "Mismatched braces (pos: " << idx << " )" << endl;
            exit(-1);
        }

        tokenstring = statement[idx];
        if(statement[idx] == '(')
            level++;
        else if(statement[idx] == ')')
            level--;
        idx++;
   }
   else if(isdigit(statement[idx]))
   {
        tokentype = DIGIT;
        tokenstring = statement[idx++];
        
        while(isdigit(statement[idx]) || ishex(statement[idx])) 
            tokenstring << statement[idx++];
   }
   else if(isalpha(statement[idx]))
   {
        tokentype = VARIABLE;
        tokenstring = statement[idx++];

        while(isalpha(statement[idx]) || statement[idx] == '.')
            tokenstring << statement[idx++];
   }
   else 
   {
        cout << "Not a legitimate character: '" << statement[idx] << "' @ " << idx << endl;
        exit(-1);
   }
   
   token = tokenstring.ptr();   
}

#define Nexttoken nexttoken()

unsigned long get_esp()
{
    asm("movl %esp, %eax");
}

void GetOperand(unsigned long *ret)
{
    if(tokentype == OPERATOR && token[0] == '(')
    {
        level++;
        DoSequence(ret);
        
        if(token[0] != ')')
        {
            cout << "Mismatched braces (pos: " << idx << " )" << endl;
            exit(-1);
        }
        level--;
    }
    else if(tokentype == DIGIT)
        *ret = strtoul(token,0,0);
    else if(!strcmp(token,"esp"))
        *ret = get_esp();
    else if(tokentype == VARIABLE)
        *ret = sizeof_sc(token);

    Nexttoken;
}

void Execute(char action, unsigned long *oper1, unsigned long *oper2)
{
    switch(action)
    {
        case '+': *oper1 = *oper1 + *oper2;
                  break;
                  
        case '-': *oper1 = *oper1 - *oper2;
                  break;
                  
        case '*': *oper1 = *oper1 * *oper2;
                  break;
                  
        case '/': *oper1 = *oper1 / *oper2;
                  break;

        case '%': *oper1 = *oper1 % *oper2;
                  break;

        case '^': *oper1 = *oper1 ^ *oper2;
                  break;

        case '<': *oper1 = (*oper1 < *oper2) ? *oper1 : *oper2;
                  break;
                  
        case '>': *oper1 = (*oper1 > *oper2) ? *oper1 : *oper2;
                  break;
    }
}

void GetOperator(unsigned long *ret)
{
    int x;

    if(tokentype == OPERATOR && (token[0] == '-' || token[0] == '+'))
    {
        x = token[0];
        Nexttoken;
    }

    GetOperand(ret);
    if(x == '-') 
        *ret *= -1;
}

void DoMaxMin(unsigned long *ret)
{
    unsigned long oper;
    int opera;

    GetOperator(ret);
    if(tokentype == OPERATOR && (token[0] == '<' || token[0] == '>'))
    {
        opera = token[0];
        Nexttoken;
        DoMaxMin(&oper);
        Execute(opera,ret, &oper);
    }
}

void DoUnair(unsigned long *ret)
{
    unsigned long oper;
    int opera;

    DoMaxMin(ret);    

    opera = token[0];
    while(tokentype == OPERATOR && (token[0] == '^' || token[0] == '%'))
    {
        Nexttoken;
        DoUnair(&oper);
        Execute(opera,ret, &oper);
        opera = token[0];
    }
}

void DoBinair(unsigned long *ret)
{
    unsigned long oper;
    int opera;
    
    DoUnair(ret);

    opera = token[0];
    while(tokentype == OPERATOR && (opera == '*' || opera == '/'))
    {
        Nexttoken;
        DoUnair(&oper);
        Execute(opera,ret, &oper);
        
        opera = token[0];
        if(opera == ')')
        {
            level--;
            Nexttoken;
            opera = token[0];
        }
    }
}

void DoSequence(unsigned long *ret)
{
    unsigned long oper;
    int opera;

    Nexttoken;
    DoBinair(ret);
    opera = token[0];
    while(tokentype == OPERATOR && (opera == '+' || opera == '-'))
    {
        Nexttoken;
        DoBinair(&oper);
        Execute(opera,ret, &oper);
        opera = token[0];
    }
}

int Calculate(char *string, unsigned long *ret)
{
    if((len = strlen(string)) >= 2048)
    {
        cout << "Statement is too big (>= 2048)" << endl;
        exit(-1);
    }
    
    strcpy(statement,string);
    level = idx = 0;
    
    DoSequence(ret);
    return *ret;
}
