#include "lstring.hpp"
#include <stdio.h>
#include <string.h>

lstring::lstring()
{
    if((buffer = new char[2048]) == 0)
    {
        fprintf(stderr,"Not enough memory\n");
        exit(0);
    }
    memset(buffer,0,2048);
    pos = 0;
    totsize = 2048;
}

lstring::lstring(lstring& copy)
{
    if((buffer = new char[copy.totsize]) == 0)
    {
        fprintf(stderr,"Not enough memory\n");
        exit(0);
    }
    memset(buffer,0,copy.totsize);
    
    pos = copy.pos;
    totsize = copy.totsize;
    memcpy(buffer,copy.buffer,pos);
}

lstring::lstring(int size)
{
    if((buffer = new char[size]) == 0)
    {
        fprintf(stderr,"Not enough memory\n");
        exit(0);
    }
    memset(buffer,0,size);
    
    pos = 0;
    totsize = size;
}

lstring::~lstring()
{
    delete buffer;
}

void lstring::clear()
{
    memset(buffer, 0 , totsize);
}

void lstring::increase(int n)
{
    char *tmp;

    if((tmp = new char[totsize + n]) == 0)
    {
        fprintf(stderr,"Not enough memory\n");
        exit(0);
    }
    memset(tmp,0,totsize + n);

    memcpy(tmp,buffer,pos);
    delete buffer;
    buffer = tmp;
    totsize += n;
}

unsigned char lstring::operator[](int n)
{
    if(n < 0 || n > totsize) return 0;
    return buffer[n];
}

lstring& lstring::operator+=(char *string) 
{
    int i = strlen(string);

    if(totsize - pos < i)
        increase((i - totsize) * 2);
        
    memcpy(&buffer[pos], string,i);
    pos += i;
    return *this; 
}

lstring& lstring::operator=(char *string) 
{
    int i = strlen(string);

    if(totsize < i)
        increase((i - totsize) * 2);

    memcpy(buffer, string,i);
    pos = i;
    return *this; 
}

lstring& lstring::operator=(char ch) 
{
    buffer[0] = ch;
    buffer[1] = 0;
    pos = (ch == 0)?0:1;
    return *this; 
}

lstring& lstring::operator+=(lstring& string) 
{
    int i = string.pos;

    if(totsize - pos < i)
        increase((i - totsize + pos) * 2);

    memcpy(&buffer[pos], string.buffer,i);
    pos += i;
    return *this;
}

lstring& lstring::operator+=(lstring *string) 
{
    int i = string->pos;

    if(totsize - pos < i)
        increase(i - totsize + pos);

    memcpy(&buffer[pos], string->buffer,i);
    pos += i;
    return *this;
}

lstring& lstring::operator=(lstring string) 
{
    int i = string.pos;

    if(totsize < i)
        increase((i - totsize) * 2);

    memcpy(buffer, string.buffer,i);
    pos = i;
    return *this;
}

lstring& lstring::operator<<(char x)
{
    if(!(totsize - pos))
        increase(10);
        
    if(x == '\0')
        buffer[pos] = x;
    else 
        buffer[pos++] = x;
    return *this;
}

lstring& lstring::operator<<(int x)
{
    if(!(totsize - pos))
        increase(10);

    if(x == '\0')
        buffer[pos] = x;
    else 
        buffer[pos++] = x;
    return *this;
}
