/* remotew00t 
 * Copyright (C) 2002 Andrew Griffiths
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef FORMATSTRING_H
#define FORMATSTRING_H

typedef struct formatstring_t {
    int padding;                /* How much padding is needed   */
    int stackpops;              /*          stackpops           */
    int bytes_written;          /*     many bytes already written */
    int bytes_written_running;
    unsigned long string_addr;
    int string_addr_offset;
    char *trigger;              /* what do we have to pad with to trigger */

    int getvalues_state;
    int getvalues_inner_state;
#define GETVALUES_STACKPAD 0    /* stack pops + padding */
#define GETVALUES_RETADDR 1     /* find ret addr */
#define GETVALUES_DIRECTACCESS 2 /* See if we can use direct parameter access */
    int getvalues_tmp;
    int getvalues_failed;

    int direct_parameter;
    int direct_parameter_total;

    int remote_bufsize;
    int optimise;
} formatstring_t;

typedef struct formatstring_string_t { /* take that for a name!@ */
    char *formatstr;
    int string_len; /* length of the above string */
    int address_num; /* number of addresses in the format string */
    int type;
} formatstring_string_t;


formatstring_t *formatstring_new();
void formatstring_put(formatstring_t *kill);
int formatstring_getpadding(formatstring_t *format, int write_byte);
void formatstring_set_trigger(formatstring_t *format, char *trigger);
char *formatstring_getvalues(formatstring_t *format, u_int32_t host, short int port);
int formatstring_analyse(formatstring_t *format, char *reply);
int formatstring_analyse_failed(formatstring_t *format);
void formatstring_set_written(formatstring_t *format, int written);

formatstring_string_t *formatstring_makeformatstring(formatstring_t 
*format, unsigned long *retloc, unsigned long retaddr);
void formatstring_string_put(formatstring_string_t *fmtstr);

void formatstring_set_optimise(formatstring_t *format, int optimise);
void formatstring_set_buffersize(formatstring_t *format, int bufsize);
void formatstring_turnoff_directparameter(formatstring_t *format);
void formatstring_mod_stackpops(formatstring_t *format, int amount);

#define WRITEFMTADDR(addr, val) {\
    (addr)[0] = (val) & 0xff;\
    (addr)[1] = (val >> 8) & 0xff;\
    (addr)[2] = (val >> 16) & 0xff;\
    (addr)[3] = (val >> 24) & 0xff;\
}

#endif /* FORMATSTRING_H */
