#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/wait.h>

#include <readline/readline.h>

int config_fd;
/* assembly required... do a shellcode to dump config_fd ;) */


void dropprivs()
{
	assert(setresgid(getgid(), getgid(), getgid()) == 0);
	assert(setresuid(getuid(), getuid(), getuid()) == 0);
}

void raiseprivs()
{
	assert(setresgid(getegid(), getegid(), getegid()) == 0);
	assert(setresuid(geteuid(), geteuid(), geteuid()) == 0);
}

void open_conf()
{
	config_fd = open("./level7_pass", O_RDONLY);
	assert(config_fd > 0);
}

void reaper(int signo)
{
	int ret, status;

	ret = wait(&status);
	if(ret == -1) return;
	if(WIFEXITED(status)) {
		printf("Proccess %d exited %s\n", ret, WEXITSTATUS(status)?"with errors":"successfully");
	} else if(WIFSIGNALED(status)) {
		printf("Proccess %d received %s\n", ret, strsignal(WTERMSIG(status)));
	}
	
}

char *read_until_char(int fd, char wot)
{
	char *ret;
	int len;

	ret = NULL;
	len = 0;
	
	while(1) {
		ret = realloc(ret, len + 2);
		assert(ret != NULL);
		ret[len + 1] = '\x00';
		
		if(read(fd, &ret[len], 1) != 1) {
			printf("?!\n");
			exit(EXIT_FAILURE);
		}
		// printf("[%c:%s]", ret[len], ret);
		if(ret[len] == wot) break;	
		len++;
	}
	//printf("\n");
	return ret;
}

void wra(int fd) 
{
	char buf[512], c, *fname, *hdr;
	int request_type; /* 0 = GET, 1 = PUT, 2 = POST */
	int major, minor, len;
	unsigned char *data;
	
	request_type = len = 0;
	memset(buf, 0, sizeof(buf));
	
	if(read(fd, buf, 4) != 4) {
		printf("\rCompletely invalid request.\n");
		exit(EXIT_FAILURE);
	}

	if(strncmp(buf, "PUT", 3) == 0) {
		request_type = 1;
	} else if(strncmp(buf, "POST", 4) ==0) {
		request_type = 2;
		if(read(fd, &c, 1) != 1) {
			printf("\r?!\n");
			exit(EXIT_FAILURE);
		}
	} else if(strncmp(buf, "GET", 3) == 0) {
		request_type = 0;
	} else {
		printf("\rDoesn't appear to be either PUT, POST or GET\n");
		exit(EXIT_FAILURE);
	}

	fname = read_until_char(fd, ' ');
	
	printf("\r%s'ing %s\n", buf, fname);
	memset(buf, 0, sizeof(buf));
	free(fname);	
	
	if(read(fd, buf, 9) != 9) {
		printf("Reading failed\n");
		exit(EXIT_FAILURE);
	}
	
	if(sscanf(buf, "HTTP/%d.%d\n", &major, &minor) != 2) {
		printf("Invalid format ('%s') for HTTP request\n");
		exit(EXIT_FAILURE);
	}

	printf("\tUsing HTTP/%d.%d\n", major, minor);

	while(1) {
		hdr = read_until_char(fd, '\n');
		if(hdr[0] == '\n') break;
		
		if(strncmp(hdr, "Content-Length:", 15)==0) {
			if(request_type == 0) {
				printf("\tWith an invalid content-length: header\n");
				printf("\tInvalid because its GET, not POST or PUT\n");
				exit(EXIT_FAILURE);
			}
			len = strtod(hdr + 16, NULL);
		}
		free(hdr);	
	}
	
	if(len) {
		printf("\tExpected data uploaded [%u:%d] bytes\n", len, len);
		data = malloc(len + 3); /* + \r\n\n */
		assert(data != NULL);
		
		if(read(fd, data, len) != len) {
			printf("Incorrect amount of bytes uploaded\n");
			free(data);
			exit(EXIT_FAILURE);
		}
	}

	printf("Everything looks intact\n");
	exit(EXIT_SUCCESS);
}

void wrh(int port) /* web request handler */
{
	int sfd;
	struct sockaddr_in sin;

	dropprivs();
	
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = inet_addr("127.0.0.1");
	sin.sin_port = htons(port);

	if((sfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		printf("\nUnable to socket(): %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	if(bind(sfd, (void *)&sin, sizeof(struct sockaddr_in)) == -1) {
		printf("\nUnable to bind(): %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	if(listen(sfd, 5) == -1) {
		printf("\nUnable to listen(): %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	signal(SIGCHLD, reaper);
	
	while(1) {
		int chld_fd;
		chld_fd = accept(sfd, NULL, 0);
		if(fork() == 0) { close(sfd); wra(chld_fd); exit(-1); }
		close(chld_fd);
	}
}

void help()
{
	printf("Web Request Online Help\n");
	printf("-----------------------\n\n");
	printf("enable\t- becomes priviledged (more options), format enable pass\n");
	printf("\twhere pass is your password specified in the config file\n");
	printf("\tand where the cmd is a valid command (ERROR: none defined)\n");
	printf("port\t- specifies which port the analyser should listen on\n");
	printf("listen\t- starts the web request helper\n");
	printf("stop\t- stops the web request helper\n\n");
	
}

int checkpass(char *pass)
{
	char pw[20];

	lseek(config_fd, 0, SEEK_SET);
	memset(pw, 0, sizeof(pw));
	if(read(config_fd, pw, sizeof(pw)) == -1) {
		printf("pw reading failed\n");
		exit(EXIT_FAILURE);
	}

	if(strcmp(pw, pass) == 0) return 1;
	
	memset(pw, 0, sizeof(pw));
	strncpy(pw, pass, sizeof(pw)-1);
	return 0;
}

int main(int argc, char **argv)
{
	int port;
	int running;
	int pid;
	int disabled;

	open_conf();

	running = port = pid = disabled = 0;
	
	printf("web-request analyser\n");
	printf("--------------------\n");
	printf("Type help to get started.\n\n");	
	
	signal(SIGCHLD, SIG_IGN);
	
	while(1) {
		char *cmd;

		cmd = readline("wra> ");
		assert(cmd);

		if(strncmp(cmd, "enable", 6)==0) {
			if(checkpass(cmd + 7) && disabled==0) {
				char *x, *y;
				
				x = malloc(8);
				y = malloc(8);

				if(read(0, x, 13) != 13) exit(EXIT_FAILURE);
			} else {
				
				printf("\rFailed authentication\n");
				dropprivs();
			}
		} else if((strncmp(cmd, "quit", 4) == 0) || (strncmp(cmd, "exit", 4)==0)) {
			exit(EXIT_FAILURE);
		} else if((strncmp(cmd, "port", 4) == 0)) {
			port = strtod(cmd + 5, NULL);
			if(port == 0) { printf("wtf are you smoking?!\n"); continue; }
		} else if(strncmp(cmd, "help", 4) == 0) {
			help();
		} else if(strncmp(cmd, "listen", 6) == 0) {
			if(!pid) if((pid = fork()) == 0) wrh(port);
		} else if(strncmp(cmd, "stop", 4) == 0) {
			kill(9, pid);
			pid = 0;
		}

	}	
	
}
